#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

__all__ = ["to_pandas"]

from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Tuple,
    Union,
)

import pandas as pd

from neptune.internal.backends.api_model import (
    AttributeType,
    AttributeWithProperties,
    LeaderboardEntry,
)
from neptune.internal.utils.logger import get_logger
from neptune.internal.utils.run_state import RunState

if TYPE_CHECKING:
    from neptune.table import Table

logger = get_logger()


def to_pandas(table: Table) -> pd.DataFrame:
    def make_attribute_value(attribute: AttributeWithProperties) -> Any:
        _type = attribute.type
        _properties = attribute.properties
        if _type == AttributeType.RUN_STATE:
            return RunState.from_api(_properties.get("value")).value
        if _type in (
            AttributeType.FLOAT,
            AttributeType.INT,
            AttributeType.BOOL,
            AttributeType.STRING,
            AttributeType.DATETIME,
        ):
            return _properties.get("value")
        if _type == AttributeType.FLOAT_SERIES:
            return _properties.get("last")
        if _type == AttributeType.STRING_SERIES:
            return _properties.get("last")
        if _type == AttributeType.IMAGE_SERIES:
            return None
        if _type == AttributeType.FILE or _type == AttributeType.FILE_SET:
            return None
        if _type == AttributeType.STRING_SET:
            return ",".join(_properties.get("values"))
        if _type == AttributeType.GIT_REF:
            return _properties.get("commit", {}).get("commitId")
        if _type == AttributeType.NOTEBOOK_REF:
            return _properties.get("notebookName")
        if _type == AttributeType.ARTIFACT:
            return _properties.get("hash")
        logger.error(
            "Attribute type %s not supported in this version, yielding None. Recommended client upgrade.",
            _type,
        )
        return None

    def make_row(entry: LeaderboardEntry) -> Dict[str, Any]:
        row: Dict[str, Union[str, float, datetime]] = dict()
        for attr in entry.attributes:
            value = make_attribute_value(attr)
            if value is not None:
                row[attr.path] = value
        return row

    def sort_key(attr: str) -> Tuple[int, str]:
        domain = attr.split("/")[0]
        if domain == "sys":
            return 0, attr
        if domain == "monitoring":
            return 2, attr
        return 1, attr

    rows = dict((n, make_row(entry)) for (n, entry) in enumerate(table._entries))

    df = pd.DataFrame.from_dict(data=rows, orient="index")
    df = df.reindex(sorted(df.columns, key=sort_key), axis="columns")

    return df



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "Signal",
    "SignalsVisitor",
    "BatchStartedSignal",
    "BatchProcessedSignal",
    "BatchLagSignal",
]

from abc import abstractmethod
from dataclasses import dataclass


@dataclass
class Signal:
    occured_at: float

    @abstractmethod
    def accept(self, visitor: "SignalsVisitor") -> None:
        ...


@dataclass
class BatchStartedSignal(Signal):
    def accept(self, visitor: "SignalsVisitor") -> None:
        visitor.visit_batch_started(signal=self)


@dataclass
class BatchProcessedSignal(Signal):
    def accept(self, visitor: "SignalsVisitor") -> None:
        visitor.visit_batch_processed(signal=self)


@dataclass
class BatchLagSignal(Signal):
    lag: float

    def accept(self, visitor: "SignalsVisitor") -> None:
        visitor.visit_batch_lag(signal=self)


class SignalsVisitor:
    @abstractmethod
    def visit_batch_started(self, signal: Signal) -> None:
        ...

    @abstractmethod
    def visit_batch_processed(self, signal: Signal) -> None:
        ...

    @abstractmethod
    def visit_batch_lag(self, signal: Signal) -> None:
        ...



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("AsyncOperationProcessor",)

import os
import threading
from pathlib import Path
from queue import Queue
from time import (
    monotonic,
    time,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)

from neptune.common.exceptions import NeptuneException
from neptune.common.warnings import (
    NeptuneWarning,
    warn_once,
)
from neptune.constants import ASYNC_DIRECTORY
from neptune.core.components.abstract import WithResources
from neptune.core.components.metadata_file import MetadataFile
from neptune.core.components.operation_storage import OperationStorage
from neptune.core.components.queue.disk_queue import DiskQueue
from neptune.envs import NEPTUNE_SYNC_AFTER_STOP_TIMEOUT
from neptune.exceptions import NeptuneSynchronizationAlreadyStoppedException
from neptune.internal.init.parameters import DEFAULT_STOP_TIMEOUT
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.operation_logger import ProcessorStopLogger
from neptune.internal.operation_processors.operation_processor import OperationProcessor
from neptune.internal.operation_processors.utils import (
    common_metadata,
    get_container_full_path,
)
from neptune.internal.signals_processing.utils import (
    signal_batch_lag,
    signal_batch_processed,
    signal_batch_started,
)
from neptune.internal.threading.daemon import Daemon
from neptune.internal.utils.disk_utilization import ensure_disk_not_overutilize
from neptune.internal.utils.logger import get_logger

if TYPE_CHECKING:
    from neptune.core.components.abstract import Resource
    from neptune.internal.backends.neptune_backend import NeptuneBackend
    from neptune.internal.container_type import ContainerType
    from neptune.internal.id_formats import UniqueId
    from neptune.internal.operation_processors.operation_logger import ProcessorStopSignal
    from neptune.internal.signals_processing.signals import Signal

logger = get_logger()


serializer: Callable[[Operation], Dict[str, Any]] = lambda op: op.to_dict()


class AsyncOperationProcessor(WithResources, OperationProcessor):
    STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS = 30.0
    STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS = float(os.getenv(NEPTUNE_SYNC_AFTER_STOP_TIMEOUT, DEFAULT_STOP_TIMEOUT))

    def __init__(
        self,
        container_id: "UniqueId",
        container_type: "ContainerType",
        backend: "NeptuneBackend",
        lock: threading.RLock,
        queue: "Queue[Signal]",
        sleep_time: float = 5,
        batch_size: int = 1000,
        data_path: Optional[Path] = None,
        should_print_logs: bool = True,
    ):
        self._should_print_logs: bool = should_print_logs

        self._data_path = (
            data_path if data_path else get_container_full_path(ASYNC_DIRECTORY, container_id, container_type)
        )

        # Initialize directory
        self._data_path.mkdir(parents=True, exist_ok=True)

        self._metadata_file = MetadataFile(
            data_path=self._data_path,
            metadata=common_metadata(mode="async", container_id=container_id, container_type=container_type),
        )
        self._operation_storage = OperationStorage(data_path=self._data_path)
        self._queue = DiskQueue(
            data_path=self._data_path,
            to_dict=serializer,
            from_dict=Operation.from_dict,
            lock=lock,
        )

        self._container_id: "UniqueId" = container_id
        self._container_type: "ContainerType" = container_type
        self._backend: "NeptuneBackend" = backend
        self._batch_size: int = batch_size
        self._last_version: int = 0
        self._consumed_version: int = 0
        self._consumer: Daemon = self.ConsumerThread(self, sleep_time, batch_size)
        self._lock: threading.RLock = lock
        self._signals_queue: "Queue[Signal]" = queue
        self._accepts_operations: bool = True

        # Caller is responsible for taking this lock
        self._waiting_cond = threading.Condition(lock=lock)

    @property
    def operation_storage(self) -> "OperationStorage":
        return self._operation_storage

    @property
    def data_path(self) -> Path:
        return self._data_path

    @property
    def resources(self) -> Tuple["Resource", ...]:
        return self._metadata_file, self._operation_storage, self._queue

    @ensure_disk_not_overutilize
    def enqueue_operation(self, op: Operation, *, wait: bool) -> None:
        if not self._accepts_operations:
            warn_once("Not accepting operations", exception=NeptuneWarning)
            return

        self._last_version = self._queue.put(op)

        if self._check_queue_size():
            self._consumer.wake_up()
        if wait:
            self.wait()

    def start(self) -> None:
        self._consumer.start()

    def pause(self) -> None:
        self._consumer.pause()
        self.flush()

    def resume(self) -> None:
        self._consumer.resume()

    def wait(self) -> None:
        self.flush()
        waiting_for_version = self._last_version
        self._consumer.wake_up()

        # Probably reentering lock just for sure
        with self._waiting_cond:
            self._waiting_cond.wait_for(
                lambda: self._consumed_version >= waiting_for_version or not self._consumer.is_running()
            )
        if not self._consumer.is_running():
            raise NeptuneSynchronizationAlreadyStoppedException()

    def _check_queue_size(self) -> bool:
        return self._queue.size() > self._batch_size / 2

    def _wait_for_queue_empty(
        self,
        initial_queue_size: int,
        seconds: Optional[float],
        signal_queue: Optional["Queue[ProcessorStopSignal]"] = None,
    ) -> None:
        waiting_start: float = monotonic()
        time_elapsed: float = 0.0
        max_reconnect_wait_time: float = self.STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS if seconds is None else seconds
        op_logger = ProcessorStopLogger(
            processor_id=id(self),
            signal_queue=signal_queue,
            logger=logger,
            should_print_logs=self._should_print_logs,
        )
        if initial_queue_size > 0:
            if self._consumer.last_backoff_time > 0:
                op_logger.log_connection_interruption(max_reconnect_wait_time)
            else:
                op_logger.log_remaining_operations(size_remaining=initial_queue_size)

        while True:
            if seconds is None:
                if self._consumer.last_backoff_time == 0:
                    # reset `waiting_start` on successful action
                    waiting_start = monotonic()
                wait_time = self.STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS
            else:
                wait_time = max(
                    min(
                        seconds - time_elapsed,
                        self.STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS,
                    ),
                    0.0,
                )
            self._queue.wait_for_empty(wait_time)
            size_remaining = self._queue.size()
            already_synced = initial_queue_size - size_remaining
            already_synced_proc = (already_synced / initial_queue_size) * 100 if initial_queue_size else 100
            if size_remaining == 0:
                op_logger.log_success(ops_synced=initial_queue_size)
                return

            time_elapsed = monotonic() - waiting_start
            if self._consumer.last_backoff_time > 0 and time_elapsed >= max_reconnect_wait_time:

                op_logger.log_reconnect_failure(
                    max_reconnect_wait_time=max_reconnect_wait_time,
                    size_remaining=size_remaining,
                )
                return

            if seconds is not None and wait_time == 0:
                op_logger.log_sync_failure(seconds=seconds, size_remaining=size_remaining)
                return

            if not self._consumer.is_running():
                exception = NeptuneSynchronizationAlreadyStoppedException()
                logger.warning(str(exception))
                return

            op_logger.log_still_waiting(
                size_remaining=size_remaining,
                already_synced=already_synced,
                already_synced_proc=already_synced_proc,
            )

    def stop(
        self, seconds: Optional[float] = None, signal_queue: Optional["Queue[ProcessorStopSignal]"] = None
    ) -> None:
        ts = time()
        self.flush()
        if self._consumer.is_running():
            self._consumer.disable_sleep()
            self._consumer.wake_up()
            self._wait_for_queue_empty(
                initial_queue_size=self._queue.size(),
                seconds=seconds,
                signal_queue=signal_queue,
            )
            self._consumer.interrupt()
        sec_left = None if seconds is None else seconds - (time() - ts)
        self._consumer.join(sec_left)

        # Close resources
        self.close()

        # Remove local files
        if self._queue.is_empty():
            self.cleanup()

    def cleanup(self) -> None:
        super().cleanup()
        try:
            self._data_path.rmdir()
        except OSError:
            pass

    def close(self) -> None:
        self._accepts_operations = False
        super().close()

    class ConsumerThread(Daemon):
        def __init__(
            self,
            processor: "AsyncOperationProcessor",
            sleep_time: float,
            batch_size: int,
        ):
            super().__init__(sleep_time=sleep_time, name="NeptuneAsyncOpProcessor")
            self._processor: "AsyncOperationProcessor" = processor
            self._batch_size: int = batch_size
            self._last_flush: float = 0.0

        def run(self) -> None:
            try:
                super().run()
            except Exception:
                with self._processor._waiting_cond:
                    self._processor._waiting_cond.notify_all()
                raise

        def work(self) -> None:
            ts = time()
            if ts - self._last_flush >= self._sleep_time:
                self._last_flush = ts
                self._processor._queue.flush()

            while True:
                batch = self._processor._queue.get_batch(self._batch_size)
                if not batch:
                    return

                signal_batch_started(queue=self._processor._signals_queue)
                self.process_batch([element.obj for element in batch], batch[-1].ver, batch[-1].at)

        # WARNING: Be careful when changing this function. It is used in the experimental package
        def _handle_errors(self, errors: List[NeptuneException]) -> None:
            for error in errors:
                logger.error(
                    "Error occurred during asynchronous operation processing: %s",
                    error,
                )

        @Daemon.ConnectionRetryWrapper(
            kill_message=(
                "Killing Neptune asynchronous thread. All data is safe on disk and can be later"
                " synced manually using `neptune sync` command."
            )
        )
        def process_batch(self, batch: List[Operation], version: int, occurred_at: Optional[float] = None) -> None:
            if occurred_at is not None:
                signal_batch_lag(queue=self._processor._signals_queue, lag=time() - occurred_at)

            expected_count = len(batch)
            version_to_ack = version - expected_count
            while True:
                # TODO: Handle Metadata errors
                processed_count, errors = self._processor._backend.execute_operations(
                    container_id=self._processor._container_id,
                    container_type=self._processor._container_type,
                    operations=batch,
                    operation_storage=self._processor._operation_storage,
                )

                signal_batch_processed(queue=self._processor._signals_queue)
                version_to_ack += processed_count
                batch = batch[processed_count:]

                with self._processor._waiting_cond:
                    self._processor._queue.ack(version_to_ack)

                    self._handle_errors(errors)

                    self._processor._consumed_version = version_to_ack

                    if version_to_ack == version:
                        self._processor._waiting_cond.notify_all()
                        return



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["common_metadata", "get_container_full_path", "get_container_dir"]

import os
import platform
import random
import string
import sys
from datetime import datetime
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
)

from neptune.constants import NEPTUNE_DATA_DIRECTORY
from neptune.metadata_containers.structure_version import StructureVersion

if TYPE_CHECKING:
    from neptune.internal.container_type import ContainerType
    from neptune.internal.id_formats import UniqueId


RANDOM_KEY_LENGTH = 8


def get_neptune_version() -> str:
    from neptune.version import __version__ as neptune_version

    return neptune_version


def common_metadata(mode: str, container_id: "UniqueId", container_type: "ContainerType") -> Dict[str, Any]:
    return {
        "mode": mode,
        "containerId": container_id,
        "containerType": container_type,
        "structureVersion": StructureVersion.DIRECT_DIRECTORY.value,
        "os": platform.platform(),
        "pythonVersion": sys.version,
        "neptuneClientVersion": get_neptune_version(),
        "createdAt": datetime.utcnow().isoformat(),
    }


def get_container_dir(container_id: "UniqueId", container_type: "ContainerType") -> str:
    return f"{container_type.value}__{container_id}__{os.getpid()}__{random_key(RANDOM_KEY_LENGTH)}"


def get_container_full_path(type_dir: str, container_id: "UniqueId", container_type: "ContainerType") -> Path:
    neptune_data_dir = Path(os.getenv("NEPTUNE_DATA_DIRECTORY", NEPTUNE_DATA_DIRECTORY))
    return neptune_data_dir / type_dir / get_container_dir(container_id=container_id, container_type=container_type)


def random_key(length: int) -> str:
    characters = string.ascii_lowercase + string.digits
    return "".join(random.choice(characters) for _ in range(length))



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OperationStorage"]

import os
import shutil
from pathlib import Path

from neptune.core.components.abstract import Resource

UPLOAD_PATH: str = "upload_path"


class OperationStorage(Resource):
    def __init__(self, data_path: Path):
        self._data_path = data_path

        # initialize upload directory
        os.makedirs(self.upload_path, exist_ok=True)

    @property
    def data_path(self) -> Path:
        return self._data_path

    @property
    def upload_path(self) -> Path:
        return self._data_path / UPLOAD_PATH

    def cleanup(self) -> None:
        shutil.rmtree(self.upload_path, ignore_errors=True)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "Container",
    "ExecutionDirectory",
    "OfflineContainer",
    "AsyncContainer",
]

import os
import shutil
import threading
import time
from abc import (
    ABC,
    abstractmethod,
)
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
)

from neptune.cli.utils import get_qualified_name
from neptune.common.exceptions import NeptuneConnectionLostException
from neptune.constants import ASYNC_DIRECTORY
from neptune.core.components.operation_storage import OperationStorage
from neptune.core.components.queue.disk_queue import DiskQueue
from neptune.envs import NEPTUNE_SYNC_BATCH_TIMEOUT_ENV
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import UniqueId
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.utils import get_container_dir
from neptune.internal.utils.logger import get_logger
from neptune.metadata_containers.structure_version import StructureVersion

if TYPE_CHECKING:
    from neptune.internal.backends.api_model import (
        ApiExperiment,
        Project,
    )
    from neptune.internal.backends.neptune_backend import NeptuneBackend


logger = get_logger(with_prefix=False)
retries_timeout = int(os.getenv(NEPTUNE_SYNC_BATCH_TIMEOUT_ENV, "3600"))


class ExecutionDirectory:
    def __init__(self, path: Path, synced: bool, structure_version: StructureVersion, parent: Optional[Path] = None):
        self._path = path
        self._synced = synced
        self._structure_version = structure_version
        self._parent = parent

    @property
    def path(self) -> Path:
        return self._path

    @property
    def synced(self) -> bool:
        return self._synced

    @property
    def structure_version(self) -> StructureVersion:
        return self._structure_version

    @property
    def parent(self) -> Optional[Path]:
        return self._parent

    def clear(self) -> None:
        if self.path.exists():
            remove_directory_structure(self.path)

    def sync(self, *, backend: "NeptuneBackend", container_id: UniqueId, container_type: ContainerType) -> None:
        operation_storage = OperationStorage(self.path)
        serializer: Callable[[Operation], Dict[str, Any]] = lambda op: op.to_dict()

        with DiskQueue(
            data_path=self.path,
            to_dict=serializer,
            from_dict=Operation.from_dict,
            lock=threading.RLock(),
        ) as disk_queue:
            while True:
                raw_batch = disk_queue.get_batch(1000)
                if not raw_batch:
                    break
                version = raw_batch[-1].ver
                batch = [element.obj for element in raw_batch]

                start_time = time.monotonic()
                expected_count = len(batch)
                version_to_ack = version - expected_count
                while True:
                    try:
                        processed_count, _ = backend.execute_operations(
                            container_id=container_id,
                            container_type=container_type,
                            operations=batch,
                            operation_storage=operation_storage,
                        )
                        version_to_ack += processed_count
                        batch = batch[processed_count:]
                        disk_queue.ack(version)
                        if version_to_ack == version:
                            break
                    except NeptuneConnectionLostException as ex:
                        if time.monotonic() - start_time > retries_timeout:
                            raise ex
                        logger.warning(
                            "Experiencing connection interruptions."
                            " Will try to reestablish communication with Neptune."
                            " Internal exception was: %s",
                            ex.cause.__class__.__name__,
                        )

    def move(self, *, base_path: Path, target_container_id: UniqueId, container_type: ContainerType) -> None:
        new_online_dir = get_container_dir(container_id=target_container_id, container_type=container_type)
        try:
            (base_path / ASYNC_DIRECTORY).mkdir(parents=True, exist_ok=True)
        except OSError:
            logger.warning(f"Cannot create directory: {base_path / ASYNC_DIRECTORY}")
            return

        self.path.rename(base_path / ASYNC_DIRECTORY / new_online_dir)
        self._path = base_path / ASYNC_DIRECTORY / new_online_dir
        self._structure_version = StructureVersion.DIRECT_DIRECTORY


class Container(ABC):
    @property
    @abstractmethod
    def container_id(self) -> UniqueId:
        ...

    @property
    @abstractmethod
    def execution_dirs(self) -> List[ExecutionDirectory]:
        ...

    @property
    def synced(self) -> bool:
        return all(map(lambda execution_dir: execution_dir.synced, self.execution_dirs))

    @abstractmethod
    def sync(self, *, base_path: Path, backend: "NeptuneBackend", project: Optional["Project"] = None) -> None:
        ...

    def clear(self) -> None:
        for execution_dir in self.execution_dirs:
            execution_dir.clear()

        for execution_dir in self.execution_dirs:
            if execution_dir.parent is not None:
                remove_directory(execution_dir.parent)
                break


class AsyncContainer(Container):
    def __init__(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        execution_dirs: List[ExecutionDirectory],
        found: bool,
        experiment: Optional["ApiExperiment"] = None,
    ):
        self._container_id = container_id
        self._container_type = container_type
        self._execution_dirs = execution_dirs
        self._found = found
        self._experiment = experiment

    @property
    def container_id(self) -> UniqueId:
        return self._container_id

    @property
    def container_type(self) -> ContainerType:
        return self._container_type

    @property
    def execution_dirs(self) -> List[ExecutionDirectory]:
        return self._execution_dirs

    @property
    def found(self) -> bool:
        return self._found

    @property
    def experiment(self) -> Optional["ApiExperiment"]:
        return self._experiment

    def sync(self, *, base_path: Path, backend: "NeptuneBackend", project: Optional["Project"] = None) -> None:
        assert self.experiment is not None  # mypy fix

        qualified_container_name = get_qualified_name(self.experiment)
        logger.info("Synchronising %s", qualified_container_name)

        for execution_dir in self.execution_dirs:
            if not execution_dir.synced:
                execution_dir.sync(
                    backend=backend,
                    container_id=self.container_id,
                    container_type=self.container_type,
                )

        self.clear()
        logger.info("Synchronization of %s %s completed.", self.experiment.type.value, qualified_container_name)


class OfflineContainer(Container):
    def __init__(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        execution_dirs: List[ExecutionDirectory],
        found: bool,
    ):
        self._container_id = container_id
        self._container_type = container_type
        self._execution_dirs = execution_dirs
        self._found = found

    @property
    def container_id(self) -> UniqueId:
        return self._container_id

    @property
    def container_type(self) -> ContainerType:
        return self._container_type

    @property
    def execution_dirs(self) -> List[ExecutionDirectory]:
        return self._execution_dirs

    @property
    def found(self) -> bool:
        return self._found

    def sync(self, *, base_path: Path, backend: "NeptuneBackend", project: Optional["Project"] = None) -> None:
        assert project is not None  # mypy fix

        experiment = register_offline_container(
            backend=backend,
            project=project,
            container_type=self.container_type,
        )

        if experiment:
            self._container_id = experiment.id
            for execution_dir in self.execution_dirs:
                execution_dir.move(
                    base_path=base_path,
                    target_container_id=self.container_id,
                    container_type=self.container_type,
                )
        else:
            logger.warning("Cannot register offline container %s", self.container_id)
            return

        qualified_container_name = get_qualified_name(experiment)
        logger.info("Offline container %s registered as %s", self.container_id, qualified_container_name)
        logger.info("Synchronising %s", qualified_container_name)

        for execution_dir in self.execution_dirs:
            execution_dir.sync(
                backend=backend,
                container_id=self.container_id,
                container_type=self.container_type,
            )

        self.clear()
        logger.info("Synchronization of %s %s completed.", experiment.type.value, qualified_container_name)


def remove_directory(path: Path) -> None:
    try:
        path.rmdir()
        logger.info(f"Deleted: {path}")
    except OSError:
        logger.warning(f"Cannot remove directory: {path}")


def remove_directory_structure(path: Path) -> None:
    try:
        shutil.rmtree(path)
        logger.info(f"Deleted: {path}")
    except OSError:
        logger.warning(f"Cannot remove directory: {path}")


def register_offline_container(
    *, backend: "NeptuneBackend", project: "Project", container_type: ContainerType
) -> Optional["ApiExperiment"]:
    try:
        if container_type == ContainerType.RUN:
            return backend.create_run(project.id)
        else:
            raise ValueError("Only runs are supported in offline mode")
    except Exception as e:
        logger.warning(
            "Exception occurred while trying to create a run" " on the Neptune server. Please try again later",
        )
        logger.exception(e)
        return None



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["FileEntry"]

import datetime
from dataclasses import dataclass
from typing import Any


@dataclass
class FileEntry:
    name: str
    size: int
    mtime: datetime.datetime
    file_type: str

    @classmethod
    def from_dto(cls, file_dto: Any) -> "FileEntry":
        return cls(name=file_dto.name, size=file_dto.size, mtime=file_dto.mtime, file_type=file_dto.fileType)
